(princ "Version 0.15, 21.8.98")
(provide "ll-macs")


(defmacro make-model ; makes a model object
    (name &key (data-table nil) (title "")
       (terms nil))
   `(progn
    (send (eval ,data-table) :add-model ',name)
    (def ,name 
      (send model-proto :new
      	:data-table ',data-table
      	:title ,title
      	:terms (eval (with-input-from-string (s 
      	  (concatenate 'string "`(",terms")")
      	                               ) (read s)))
      )
    )
    (send ,name :needs-computing t)
    (setf *current-model* ',name))
)

(defmacro fit (&optional (model-object *current-model*)
                     &key (iterations 10)) 
; just a calling macro
	`(fit-function ',model-object :iterations ,iterations))
	
	
(defmacro make-table ; makes a table object
   (name &key (title "") 
              (counts nil) 
              (factors nil) 
              (exposures nil))
    `(def ,name 
      (send table-proto :new
      	:title ,title
   		:counts ,counts
   		:factors ,factors
   		:dimensions (mapcar #'factor-levels ,factors)
   	    :exposures (if ,exposures ,exposures (repeat 1 (length ,counts))))))
   	    
(defmacro indexing-factors (&rest factors-and-levels)   
; calculates values for each named factor (with no. of levels)
	(let* ((ways (/ (length factors-and-levels) 2))
	       (factor-list (array-to-nested-list 
	         (make-array (list ways 2) :initial-contents factors-and-levels)))
	       (dimensions (mapcar #'get-levels factor-list))
		   (table-size (prod dimensions))
		   (names (mapcar #'get-name factor-list))
		   (cumulative-products (combine 1 (accumulate #'* dimensions)))
		   (temp '())
		   (values 
		     (dotimes (i ways temp)
		          (setf temp (append temp (list
		             (let* ((dimension (select dimensions i))
		               (cum-prod (select cumulative-products (+ i 1)))
		               (cum-prod-lagged (select cumulative-products i))
		              )
		              (repeat 
		                (repeat (iseq 1 dimension) 
		                  (repeat (/ table-size cum-prod) dimension))
		                cum-prod-lagged))
		            )))))
		  )
		  
		  (combine 'list (map 'list #'(lambda (name levels vals)
		     (make-factor :name name :levels levels :values vals)) 
		     names dimensions values))
		  )
)

