(princ "Version 0.15, 21.8.98")
(provide "ll-menu")

(cond ((boundp 'model-menu) (send model-menu :remove))
      (t nil))

(setf model-menu (send menu-proto :new "Model"))

; First of all a couple of useful utilities for list-items in dialogs.
; Note that the number of cells in a list item is fixed, so for example
; inserting a cell means that a cell is lost from the bottom.

(defmeth list-item-proto :remove-cell (index)
	(cond (index
	  (let ((list-data (send self :slot-value 'list-data))
         )
         (dolist (i (iseq index (- (length list-data) 1)) nil)
            (cond ((< i (- (length list-data) 1)) 
              (send self :set-text i (select list-data (+ i 1))))
                  (t (send self :set-text i "")))
               )))
          (t nil)))
               
(defmeth list-item-proto :insert-cell (index text)
	(cond (index
	  (let* ((list-data (send self :slot-value 'list-data))
		 )
		 (dolist (i (iseq (- (length list-data) 1) (+ index 1)) nil)
		 	(send self :set-text i (select list-data (- i 1))))
         (send self :set-text index text)
          ))
          (t nil)))
          
(setf make-new-model-item (send menu-item-proto :new "New..."
 	:action #'(lambda ()
 	   (cond ((mapcar #'tablep (variables)) ; at least one table exists
 	    (let* ((current-model-exists (boundp '*current-model*))
 	          (list-of-tables (select (variables) 
                (which (mapcar #'tablep (variables)))))
              (old-current-table 
 	           (if current-model-exists
 	             (send (eval *current-model*) :data-table)
 	              (first list-of-tables)))
              (index-of-old-current-table 
                 (if current-model-exists (position old-current-table list-of-tables) 0))
 	          (table-info-item (send text-item-proto :new (table-info old-current-table) 
	              :text-length 50))
	          (header-item (send text-item-proto :new "NEW MODEL SPECIFICATION"))
 	          (table-list-item (send list-item-proto :new 
                (mapcar #'symbol-name list-of-tables)
                  :action #'(lambda (x) 
                       (let* ((table-name (select list-of-tables
	                                   (send table-list-item :selection)))
	                         )
                       (send table-info-item :text (table-info table-name))))))
	          (table-prompt-item (send text-item-proto :new "Table to be modelled:"))
              (fill-item (send text-item-proto :new " "))
              (terms-prompt-item (send text-item-proto :new 
                "Model terms:"))
              (terms-item (send edit-text-item-proto :new ; no longer used
                    (if current-model-exists          ; (see terms-list-item below)
                     (strip-first-and-last
                      (with-output-to-string (s) 
                      (prin1 
                         (send (eval *current-model*) :terms)
                           s)))
                      "")))
              (terms-list-item (send list-item-proto :new 
                    (if current-model-exists
                      (combine (mapcar #'(lambda (term) (with-output-to-string (s) 
                      (prin1 term s)))
                         (send (eval *current-model*) :terms)) (repeat "" 90))
                      (repeat "" 10)
                     )))
              (delete-term-button (send button-item-proto :new "Delete term"
                 :action 
                    #'(lambda () 
                    (let* ((selection (send terms-list-item :selection))
                          ) 
                      (send terms-list-item :remove-cell selection)
                     ))))
              (insert-term-button (send button-item-proto :new "Insert term"
                 :action #'(lambda () 
                     (let* ((number-of-terms 
                               (if current-model-exists 
                                  (length (send (eval *current-model*) :terms))
                                  0))
                            (selection (if (send terms-list-item :selection)
                                           (send terms-list-item :selection)
                                           (position "" 
                                            (send terms-list-item :slot-value 'list-data)
                                             :test 'equalp)))
                            (new-term-fixed-text (send text-item-proto :new 
                              "New term to insert:"))
                            (new-term-text-item (send edit-text-item-proto :new "(  )"
                              :text-length 35))
                            (ok-button (send modal-button-proto :new "OK"
                              :action #'(lambda ()
                                 (send terms-list-item :insert-cell selection 
                                    (send new-term-text-item :text))
                                 )))
                            (cancel-button (send modal-button-proto :new "Cancel"
                               ))
                            (new-term-dialog (send modal-dialog-proto :new
                             (list
                             new-term-fixed-text new-term-text-item ok-button cancel-button)
                              :default-button ok-button))
                           )
                          (send new-term-dialog :modal-dialog)
                     ))))
              (edit-term-button (send button-item-proto :new "Edit term"
                 :action #'(lambda () 
                   (cond ((send terms-list-item :selection)
                   (let* ((selection (send terms-list-item :selection))
                          (edit-term-fixed-text (send text-item-proto :new 
                              "Edited term:"))
                            (text (select (send terms-list-item :slot-value 'list-data)
                              selection))
                            (edit-term-text-item (send edit-text-item-proto :new 
                              (if (equalp text "") "(  )" text)
                              :text-length 35))
                            (ok-button (send modal-button-proto :new "OK"
                              :action #'(lambda ()
                                 (setf text (send edit-term-text-item :text))
                                 )))
                            (cancel-button (send modal-button-proto :new "Cancel"
                               ))
                            (edit-term-dialog (send modal-dialog-proto :new
                             (list
                             edit-term-fixed-text edit-term-text-item ok-button cancel-button)
                              :default-button ok-button))
                           )
                          (send edit-term-dialog :modal-dialog)
                          (send terms-list-item :set-text selection  text)
                     ))
                       (t nil)))))
              (name-prompt-item (send text-item-proto 
                      :new "Model name (no spaces):"))
              (name-item (send edit-text-item-proto :new 
                  (if current-model-exists
                      (symbol-name *current-model*)
                      "")))
              (describe-prompt-item (send text-item-proto 
                      :new "Description of model (optional):"))
              (describe-item (send edit-text-item-proto :new 
                 (if current-model-exists
                   (send (eval *current-model*) :title)
                   "")))
              (dummy (send name-item :slot-value 'size (list 200 16)))
              (box-width (- (select (screen-size) 0) 40))
              (dummy (send describe-item :slot-value 'text-length 40))
              (dummy (send describe-item :slot-value 'size (list 550 25)))
              (dummy (send terms-list-item :slot-value 'size 
                 (list 250 100)))
              (ok-button (send modal-button-proto :new "OK" 
                           :action #'(lambda ()
                               (let* ((model-name (send name-item :text))
	                             (table-name (select list-of-tables
	                                   (send table-list-item :selection)))
	                             (terms (send terms-list-item :slot-value 'list-data))
	                             (terms ; remove the surplus quotes
	                               (reduce #'(lambda (string1 string2) 
	                                  (concatenate 'string string1 string2))
	                                    terms))
	                             (title (send describe-item :text))
	                           )
	                            (eval (with-input-from-string (s 
                           (concatenate 'string "(make-model " model-name " 
    						:data-table " (symbol-name table-name)
    						" :title " 
    						(with-output-to-string (s) (prin1 title s))
    						" :terms " 
    						(with-output-to-string (s) (prin1 terms s))
    						"))")) (read s)))))))
	          (cancel-button (send modal-button-proto :new "Cancel"))
              (new-model-dialog (send modal-dialog-proto :new
                (list
                  header-item fill-item
                  (list table-prompt-item  table-list-item) table-info-item fill-item
                  (list name-prompt-item name-item)
                  (list terms-prompt-item terms-list-item 
                         (list delete-term-button insert-term-button edit-term-button)) 
                  describe-prompt-item describe-item
                  fill-item
                    (list (list ok-button) (list cancel-button)))
                :default-button ok-button
                ))
            )
             (send table-list-item :selection index-of-old-current-table)
             (send new-model-dialog :modal-dialog)
       ))
       (t (terpri)
	      (princ "You must define a table first")
	      (terpri)
	      (princ "> "))
       ))))
       

       
(setf name-item (send menu-item-proto :new "Name"
		:action #'(lambda () 
		    (terpri)
		    (princ (concatenate 'string 
		      "The current model has name " (symbol-name *current-model*)
		      ", and description '" (send (eval *current-model*) :title) "'"))
		    (terpri)
		    (princ "> "))))

(setf terms-item (send menu-item-proto :new "Terms"
		:action #'(lambda () 
		    (terpri)
		    (princ (concatenate 'string 
		      "Terms in the model " (symbol-name *current-model*) " are:"))
			(terpri)
			(format t (print-without-parens (send (eval *current-model*) :terms)))
		    (terpri)
		    (princ "> "))))
		    

		       
(setf fit-item (send menu-item-proto :new "Fit"
	:action #'(lambda ()
	             (let* ((iterations (first
	                (get-value-dialog "Number of iterations:"
	                  :initial 10)))
	                   )
	             (cond (iterations
	              (fit-function *current-model* :iterations iterations))
	                   (t nil))
	             (terpri)
	             (princ "> ")
	             )
	          )
))

(setf recycle-item (send menu-item-proto :new "Recycle"
	:action #'(lambda ()
	             (cond ((send (eval *current-model*) :betas)
	                    (let* ((iterations (first
	                       (get-value-dialog "Number of further iterations:"
	                            :initial 1)))
	                          )
	                    (cond (iterations (recycle iterations))
	                          (t nil))
	                    (terpri) (princ "> ")))
	                   (t (terpri)
	                      (princ "Model not yet fitted: can't recycle")
	                      (terpri)
	                      (princ "> ")
	                   )))))

	             
(setf deviance-item (send menu-item-proto :new "Deviance"
		:action #'(lambda () 
		    (cond ((send (eval *current-model*) :needs-computing)
		             (terpri) (format t "No deviance: model ~a has not been fitted yet" 
		                  (symbol-name *current-model*)) (terpri) (princ "> "))
		          (t
		    (terpri)
		    (format t "~a~a~a~f"
		      "Deviance of model " (symbol-name *current-model*) " is: "
			    (send (eval *current-model*) :deviance))
		    (terpri)
		    (format t "~a~d"
		      "Degrees of freedom are " 
			    (send (eval *current-model*) :df))
		    (terpri)
            (princ "> "))))))
            
(setf delta-item (send menu-item-proto :new "Delta"
		:action #'(lambda () 
		    (cond ((send (eval *current-model*) :needs-computing)
		             (terpri) (format t "No delta: model ~a has not been fitted yet" 
		                  (symbol-name *current-model*)) (terpri) (princ "> "))
		          (t
		    (terpri)
		    (format t "~a~a~a~5,2f~a"
		      "Index of dissimilarity for model " (symbol-name *current-model*) " is: "
			    (send (eval *current-model*) :index-of-dissimilarity) "%")
		    (terpri)
            (princ "> "))))))
            
(setf summary-statistics-item (send menu-item-proto :new "Summary statistics"
	:action #'(lambda ()
		(cond 
		    ((send (eval *current-model*) :needs-computing)
		       (terpri) 
		       (format t "No statistics available: model ~a has not been fitted yet" 
		          (symbol-name *current-model*)) (terpri) (princ "> "))
		    (t
		    (terpri)
		    (format t "~a~a~a~f"
		      "Deviance of model " (symbol-name *current-model*) " is: "
			    (send (eval *current-model*) :deviance))
		    (terpri)
		    (format t "~a~d"
		      "Degrees of freedom: " 
			    (send (eval *current-model*) :df))
			(terpri)
		    (format t "~a~7,4f"
		      "p-value for lack of fit: "
			    (send (eval *current-model*) :p-value))
			(terpri)
		    (format t "~a~5,2f~a"
		      "Index of dissimilarity: " 
			    (send (eval *current-model*) :index-of-dissimilarity) "%")
		    (terpri)
            (princ "> "))))))

(setf betas-item (send menu-item-proto :new "Estimates"
		:action #'(lambda () 
		    (cond ((send (eval *current-model*) :needs-computing)
		             (terpri)
		             (format t "No estimates: model ~a has not been fitted yet" 
		                  (symbol-name *current-model*)) 
		                    (terpri) (princ "> "))
		          (t (terpri)
		    (princ (concatenate 'string 
		      "Parameter estimates for the model " (symbol-name *current-model*) ":"))
			(terpri)
			(print-betas (eval *current-model*))
		    (terpri)
		    (princ "> "))))	))
		    
(setf standard-errors-item (send menu-item-proto :new "Standard errors"
		:action #'(lambda () 
		    (cond ((send (eval *current-model*) :needs-computing)
		             (terpri)
		             (format t "No standard errors: model ~a has not been fitted yet" 
		                  (symbol-name *current-model*)) 
		                    (terpri) (princ "> "))
		    (t
		    (let* ((model (eval *current-model*))
		           (rows (length (send model :terms)))
		           (cols (send model :depth))
		           (term-numbers (list "Term 1:" "Term 2:" "Term 3:" "Term 4:" "Term 5:"
		             "Term 6:" "Term 7:" "Term 8:" "Term 9:" "Term 10:"
		             "Term 11:" "Term 12:" "Term 13:" "Term 14:" "Term 15:"
		             "Term 16:" "Term 17:" "Term 18:" "Term 19:" "Term 20:"
		             "Term 21:" "Term 22:" "Term 23:" "Term 24:" "Term 25:"
		             "Term 26:" "Term 27:" "Term 28:" "Term 29:" "Term 30:"))
		           (array-of-terms (bind-columns (select term-numbers (iseq 0 (- rows 1)))
		              (send model :array-of-terms)))
		           (terms-table (send list-item-proto :new 
		              array-of-terms
		                             :columns (+ 1 cols)))
		        (ok-button (send modal-button-proto :new "OK" 
                           :action #'(lambda () 
                            (let* ((effect-index (send terms-table :selection))
                                                   
                                   (term-number (select effect-index 0))
                                   (effect-name 
                                        (select array-of-terms 
                                          term-number (select effect-index 1)))
                                  )
                         (print-se (- effect-index (list 0 1)) effect-name)
                            ))))
                (cancel-button (send modal-button-proto :new "Cancel"))
                (fixed-text (send text-item-proto :new
                  "Standard errors are available for estimates of simple
multiplicative coefficients.  Select the effect of interest."))
		        (terms-dialog (send modal-dialog-proto :new
                (list
                  fixed-text terms-table ok-button cancel-button)
                ))
            )
            (send terms-dialog :modal-dialog)))))))

(setf residuals-item (send menu-item-proto :new "Residuals"
		    :action #'(lambda () 
		    (cond ((send (eval *current-model*) :needs-computing)
		             (terpri)
		             (format t "No residuals: model ~a has not been fitted yet" 
		                  (symbol-name *current-model*)) 
		                    (terpri) (princ "> "))
		          (t (terpri)
		    (princ (concatenate 'string 
		      "Standardized residuals (observed-fitted)/sqrt(fitted) 
		       for the model " (symbol-name *current-model*) ":"))
			(terpri)
			(print-residuals (eval *current-model*))
		    (terpri)
		    (princ "> "))))	))

(setf change-current-model-item (send menu-item-proto :new "Switch model"
 	:action #'(lambda ()
 	   (let* ((old-current-model *current-model*)
 	          (list-of-models (select (variables) 
                (which (mapcar #'modelp (variables)))))
              (index-of-old-current-model 
                 (position old-current-model list-of-models))
 	          (model-list-item (send list-item-proto :new 
                (mapcar #'symbol-name list-of-models)))
              (prompt-item (send text-item-proto :new 
"Choose a new
Current Model from:"))
              (fill-item (send text-item-proto :new " "))
              (ok-button (send modal-button-proto :new "OK" :action 
                   #'(lambda () 
                      (setf *current-model* (select list-of-models
                                              (send model-list-item :selection))))))
              (cancel-button (send modal-button-proto :new "Cancel"))
              (current-model-dialog (send modal-dialog-proto :new
                (list
                  (list
                    (list prompt-item model-list-item)
                    (list  fill-item ok-button cancel-button)))
                :default-button ok-button))
            )
             (send model-list-item :selection index-of-old-current-model)
             (send current-model-dialog :modal-dialog)
       ))))
       

(setf separator-item (send dash-item-proto :new))

(send model-menu :append-items 
   make-new-model-item 
   separator-item 
   name-item terms-item 
   separator-item
   fit-item
   recycle-item
   separator-item
   summary-statistics-item betas-item standard-errors-item residuals-item
   separator-item 
   change-current-model-item)

;(setf current-model-menu (send menu-proto :new "Current model"))
 
;(send current-model-menu 
;   :append-items name-item terms-item deviance-item betas-item)

;(send model-menu :append-items make-new-model-item current-model-menu 
;  change-current-model-item)
   
(setf *startup-functions* (combine *startup-functions* 'install-model-menu))






#|(defun collect-name ()
	(send current-model-name :text))|#
	
(defun install-model-menu ()
	(cond ((send model-menu :slot-value 'hardware-address); is the menu already 
	                                                      ;   installed?
	         (send model-menu :remove)
	         (send model-menu :install))
	      (t (send model-menu :install)))
)

(defun table-info (table-name)
	(concatenate 'string "  (variables "
	   (insert-commas-into
	   (strip-first-and-last
                      (with-output-to-string (s) 
                      (prin1 
                         (mapcar #'factor-name (send (eval table-name) :factors))
                           s)))) ")" ))
                           
(defun insert-commas-into (string); replaces each space with ", "
	(let* ((string-listed (coerce string 'list))
	      )
	(coerce (combine (mapcar #'(lambda (char) (if (eq char #\Space) 
	       (list #\, #\Space) char)) string-listed)) 'string)))
	       

               
	


(install-model-menu)
	   

	        
