(princ "Version 0.15, 21.8.98")
(provide "ll-modl")

(defproto model-proto '(title data-table terms subsets count-margins betas 
effects fit-values deviance df needs-computing))

(defmeth model-proto :title (&optional (title nil set))
	(if set (setf (slot-value 'title) title))
	(slot-value 'title))
	
(defmeth model-proto :data-table (&optional (data-table nil set))
	(if set (setf (slot-value 'data-table) data-table))
	(slot-value 'data-table))
	
(defmeth model-proto :terms (&optional (terms nil set))
	(if set (setf (slot-value 'terms) terms))
	(slot-value 'terms))
	
(defmeth model-proto :counts () 
    (send (eval (send self :data-table)) :counts))
    
(defmeth model-proto :factors () 
    (send (eval (send self :data-table)) :factors))
    
(defmeth model-proto :defining-factor-names () ;now redundant
; returns the names of the defining factors of the table modelled
	(mapcar #'(lambda (list) (select list 0)) 
	      (send (eval (send self :data-table)) :factors)))
	      
(defmeth model-proto :subsets (&optional (subsets nil set))
	(if set (setf (slot-value 'subsets) subsets))
	(slot-value 'subsets))
	
(defmeth model-proto :count-margins (&optional (count-margins nil set))
	(if set (setf (slot-value 'count-margins) count-margins))
	(slot-value 'count-margins))	
	
(defmeth model-proto :betas (&optional (betas nil set))
	(if set (setf (slot-value 'betas) betas))
	(slot-value 'betas))
	
(defmeth model-proto :effects (&optional (effects nil set))
(if set (setf (slot-value 'effects) effects))
(slot-value 'effects))

(defmeth model-proto :exposures ()
	(send (eval (send self :data-table)) :exposures))

(defmeth model-proto :fit-values (&optional (fit-values nil set)) 
	(if set (setf (slot-value 'fit-values) fit-values)
	   (let* ((fit-values-from-slot (slot-value 'fit-values)); possibly nil
	      )
	      (if fit-values-from-slot fit-values-from-slot
	         (send self :calculate-fit-values)))))
	         
	
(defmeth model-proto :calculate-fit-values ()	
  (setf (slot-value 'fit-values)
	(cond ((send self :betas)
	  (let* ((effects (send self :effects))
	         (exposures (send self :exposures))
	        )
	       (* exposures
	         (reduce #'* effects))))
	      (t nil))))
	      
(defmeth model-proto :deviance () (send self :calculate-deviance))
	  
(defmeth model-proto :p-value ()
   (let* ((deviance (send self :deviance))
          (df (send self :df))
         )
         (- 1 (chisq-cdf deviance df))))

(defmeth model-proto :index-of-dissimilarity ()
	(let* ((counts (send self :counts))
	       (fit (send self :fit-values))
	      )
	      (* 50 (/ (sum (abs (- counts fit))) (sum counts)))))
	  
(defmeth model-proto :calculate-deviance ()	  
	  (let* ((counts (send self :counts))
	         (fit (send self :fit-values))
	         (calculated-deviance 
	           (cond (fit (deviance counts fit))
	                 (t nil))
	        ))
	    (setf (slot-value 'deviance) calculated-deviance)))

	
(defmeth model-proto :df ()
 	(let* ((df-from-slot (slot-value 'df)); possibly nil
	      )
	(if df-from-slot df-from-slot
	  (send self :calculate-df))))

#|(defmeth model-proto :calculate-df () ; now replaced (see below)
	(setf (slot-value 'df) 
	 (cond ((send self :betas)
	   (let* ((design-matrix (send self :local-design-matrix))
	          (n (length (send (eval (send self :data-table)) :counts)))
	          (w (diagonal (send self :fit-values)))
	         )
	   (- n (matrix-rank (matmult (transpose design-matrix)  
	                      (matmult w design-matrix))))
       ))
          (t nil))
))
|#


(defmeth model-proto :calculate-df ()
	(setf (slot-value 'df) 
	 (cond ((send self :betas)
	   (let* ((design-matrix (send self :local-design-matrix))
	          (n (length (send (eval (send self :data-table)) :counts)))
	         )
	   (- n (matrix-rank (cross-product design-matrix)))
       ))
          (t nil))
))

(defmeth model-proto :depth () ; eg 2 for bilinear, etc.
	(max (mapcar #'length (send self :terms))))
	
(defun put-quotes-round (list)
 (cond ((compound-data-p (first list)) ; is the term multiplicative?
    (mapcar #'put-quotes-round list))
    (t (with-output-to-string (s) (prin1 list s)))))

(defmeth model-proto :array-of-terms () ; delivers an array of strings
	(let* ((terms (send self :terms))
	       (quoted-terms (put-quotes-round terms))
	       (depth (send self :depth))
	      )
	      (make-array (list (length terms) depth) :initial-contents
	      (mapcar #'(lambda (element) 
	         (if (stringp element) (append (list element) (repeat "" (- depth 1)))
	                (append element (repeat "" (- depth (length element)))))) 
	       quoted-terms))))

(defmeth model-proto :local-design-matrix ()
	(let* ((subsets (send self :subsets))
		   (betas (send self :betas))
		  )
		  (reduce #'bind-columns (map-elements #'design-element subsets betas))))

(defmeth model-proto :needs-computing (&optional (needs-computing nil set))
	(if set (setf (slot-value 'needs-computing) needs-computing))
	(slot-value 'needs-computing))
	
(defmeth model-proto :residuals ()
	(let* ((counts (send self :counts))
	       (fitted (send self :fit-values))
	      )
	      (/ (- counts fitted) (sqrt fitted))))
	      
(defmeth model-proto :save () ; called automatically by (savevar 'model-name)
   `(send model-proto :new
       :data-table ',(send self :data-table)
       :title  ',(send self :title)
       :terms ',(send self :terms)
       :count-margins ',(send self :count-margins)
       :subsets ',(send self :subsets)
       :betas ',(send self :betas)
       :effects ',(send self :effects)
       :fit-values ',(send self :fit-values)
       :df ',(send self :df)
       :needs-computing ',(send self :needs-computing)
))       
	      


(defun terms-are-all-factors-p (model)
; checks whether all factor names used in the "terms" expression
; are valid names of factors for the table being modelled
; -- returns T is so.
	(let* ((terms (send model :terms))
	       (factor-names (mapcar #'factor-name
	                (send (eval (send model :data-table)) :factors)))
	      )
	      (flet ((is-a-factor (term) (member term (combine factor-names))))
	          (not (member nil (mapcar #'is-a-factor (combine terms)))))))
	          

	
