(princ "Version 0.15, 21.8.98")
(provide "ll-prin")

(defun print-betas (model &key (float-digits 3))
	(let* ((terms (send model :terms))
	 	  )
	 	  (dotimes (i (length terms) (format t "~a" ""))
	 	  	(let* ((term (select terms i))
	 	  	      )
	 	  	      (format t 
	 	  	        "~%~a~d~a~a~a~%" "--- Term " (+ 1 i) ": " 
	 	  	              (string-form term) " ---")
	 	  	      (print-term-betas model i :float-digits float-digits))
	 	  )))
	 	  	
(defun print-term-betas (model term-number &key (float-digits 3)) 
	(let* ((term (select (send model :terms) term-number))
	       (term-betas (select (send model :betas) term-number))
	      )
	      (cond ((compound-data-p (first term)) 
	               (dotimes (j (length term))
	               	  (let* ((item (select term j))
	               	         (betas (select term-betas j))
	               	         
	               	        )
	               	        (print-item-betas 
	               	          model item betas :float-digits float-digits))))
	            (t 
	                (print-item-betas model term term-betas 
	                    :float-digits float-digits)))))
	                
(defun print-item-betas (model term term-betas &key (float-digits 3))
	(let* ((dimensions (dimensions model term))
	       (array-order (length dimensions))
	       (beta-array (make-array dimensions :initial-contents term-betas))
	       (last-index (- (length dimensions) 1))
	       (penultimate-index (- (length dimensions) 2))
	    )
	      (table-print beta-array term penultimate-index last-index 
	         :float-digits float-digits)))

(defun dot-strings (string1 string2)
	(concatenate 'string string1 "." string2))
	
(defun mult-strings (string1 string2)
	(concatenate 'string  string1 "*" string2 ))
	
(defun string-form (term)
; returns a single-string representation of a model term for printing
	(cond ((compound-data-p (first term)) 
	        (let* ((subterm-strings (mapcar #'string-form term)))
	              (reduce #'mult-strings subterm-strings))
	       )
		  (t (concatenate 'string 
		    "[" (reduce #'dot-strings (mapcar #'symbol-name term)) "]")
		  )
    ))
    
(defun print-residuals (model &key (float-digits 2))
	(let* ((obs (send model :counts))
		   (fit (send model :fit-values))
		   (dim (send (eval (send model :data-table)) :dimensions))
		   (ways (length dim))
		   (res (make-array dim :initial-contents (/ (- obs fit) (sqrt fit)))))
          (terpri)
          (table-print res 
            (mapcar #'factor-name (send model :factors)) 
            (- ways 2) (- ways 1)
           :float-digits float-digits)
          (values)
    )
)

(defun print-obs-and-fitted (model &key (float-digits 2))
	(let* ((obs (send model :counts))
		   (fit (send model :fit-values))
		   (dim (send (eval (send model :data-table)) :dimensions))
		   (ways (length dim))
		   (obs (make-array dim :initial-contents obs))
		   (fit (make-array dim :initial-contents fit)))
          (terpri)
          (table-print obs 
            (mapcar #'factor-name (send model :factors)) 
            (- ways 2) (- ways 1)
           :float-digits float-digits)
          (terpri)
          (table-print fit 
            (mapcar #'factor-name (send model :factors)) 
            (- ways 2) (- ways 1)
           :float-digits float-digits)
          (values)
    )
)


(defun print-se (effect-index effect-name)
	(let* ((model (eval *current-model*))
	       (term-number (select effect-index 0))
	       (term (select (send model :terms) term-number))
	      )
	      (cond ((and (multiplicative-p term) 
                                   (one-word-p effect-name))
               (let*
                 ((decimal-index (sum (* (list 10 1) 
                       effect-index)))
                  (term-columns (which (= decimal-index 
                       (index-betas (send model :betas))))) 
                  (betas (send model :betas))
	              (betas (select (combine betas) term-columns))
	              (x (send model :local-design-matrix))
	              (w (diagonal (send model :fit-values)))
	              (xtx (matmult (transpose x) w x))
	              (xtxinv (ginv xtx))
	              (b (submatrix xtxinv term-columns term-columns))
	              (standard-errors
	                 (mapcar #'(lambda (j)
                       (let* ((sigma (submatrix b (list 0 j) (list 0 j)))
                              (derivs 
                                (vector (- (/ (select betas j) 
                                          (^ (select betas 0) 2))) 
                                            (/ 1 (select betas 0))))
                             )
                             (sqrt (matmult derivs sigma derivs))))
                      (iseq (length term-columns))))
                 (standard-errors (combine 0 (rest standard-errors)))
                )
          (terpri)
          (princ (concatenate 'string
            "Estimates and standard errors for the multiplier effect " effect-name 
            " from term " (format nil "~d" (+ 1 term-number)) ":"))
          (terpri)
          (print-matrix 
            (bind-columns  (combine 1 (rest (/ betas (first betas)))) standard-errors) 
          )
         )) ; close let* and first clause of cond
         (t 
         (terpri)
         (princ 
"Standard errors are currently available only for simple factors
in multiplicative model terms, such as the association multipliers 
in a UNIDIFF model.") (terpri) (princ "> "))
         )))
         
