(princ "Version 0.15, 21.8.98")
(provide "ll-tabl")

(defstruct factor (name nil) (levels 1) (values nil))

(defun get-levels-as-vector (factor)
	(coerce (iseq 1 (factor-levels factor)) 'vector))
	
;(defun get-values (factor)
;	(select factor 2))

(defun margin (counts factors)
; calculates a list of marginal totals in list *counts*
; for all possible combinations of levels of list *factors* in fully expanded form
; (with *nil* if the combination does not occur in the table)
	(def marg-list '())
	(dolist (i 
            (ravel (reduce #'outer-list (mapcar #'get-levels-as-vector factors)))
            (rest marg-list))
	        (def marg-list 
	          (combine marg-list 
	            (sum   ;select all counts with combination i of factors
	            (select counts (which 
	                (reduce #'vectorized-and 
	                  (map-elements #'= (mapcar #'factor-values factors) i))
	                )))))))
	                             
(defproto table-proto '(title counts exposures factors dimensions models))

(defmeth table-proto :counts (&optional (counts nil set))
	(if set (setf (slot-value 'counts) counts))
	(slot-value 'counts))
	
(defmeth table-proto :exposures ()
	(slot-value 'exposures))
	
(defmeth table-proto :title (&optional (title nil set))
	(if set (setf (slot-value 'title) title))
	(slot-value 'title))
	
(defmeth table-proto :factors 
 (&optional (list-of-factors nil set))
	(if set (setf (slot-value 'factors) list-of-factors)
	        (slot-value 'factors)))

;(defmeth table-proto :factors (&optional (factor-list nil set))
;	(if set (setf (slot-value 'factors) (expand-factors factor-list)))
;	(slot-value 'factors))
	
(defmeth table-proto :add-factor (factor)
    (setf (slot-value 'factors) 
       (adjoin factor (slot-value 'factors))))
	
(defmeth table-proto :defining-factor-names () ;now redundant
; returns the names of the defining factors of a table
	(mapcar #'(lambda (list) (select list 0)) (send self :factors)))
	
(defmeth table-proto :models (&optional (models nil set))
	(if set (setf (slot-value 'models) models))
	(slot-value 'models))
	
(defmeth table-proto :add-model (model-name)
	(send self :models 
	   (remove-duplicates 
	     (non-nil (combine (send self :models) model-name)))))
	     
(defmeth table-proto :dimensions ()
	(slot-value 'dimensions))
	
#|(defun make-table ; makes a table object -- now superseded by a macro
   (&key (title nil) (counts nil) (factors nil) (dimensions nil))
   (send table-proto :new
   								:title title
   								:counts counts
   								:factors (expand-factors factors)
   								:dimensions dimensions))
|#
   								
(defun get-levels (factor)
	(select factor 1))
		   
(defun get-name (factor)
	(select factor 0))


#|
(defun extract-factor (factor-name list-of-factors) ; old version
; picks out a factor by its name
	(let* ((full-record (select list-of-factors (which 
	          (map-elements #'eq (mapcar #'first list-of-factors) factor-name)))))
	       (first full-record)))
;(compile 'extract-factor)
|#

(defun extract-factor (factor-name list-of-factors)
; picks out a factor by its name
	(first (select list-of-factors (which
	  (map-elements #'eq (mapcar #'factor-name list-of-factors) 
	     factor-name)))))
	     
