(princ "Version 0.15, 21.8.98")
(provide "ll-util")

(defun index-plot (list) (plot-points (iseq 1 (length list)) list))

(defun vec-to-list (x) (coerce x 'list))
;(compile 'vec-to-list)

(defun both (x y) (and x y)); defines and as a function
;(compile 'both)

(defun multiplicative-p (term) ; is a model term multiplicative?
   (compound-data-p (first term)))
   
(defun one-word-p (string) ; is the string free of spaces?
  (if (find #\Space string) nil t))

(defun as-binary (list-of-logicals)
	(flet ((binary (logical) (if logical 1 0)))
	   (mapcar #'binary list-of-logicals)))
;(compile 'as-binary)

(defun vectorized-and (x y) (map-elements #'both x y))
; applies "and" componentwise to a pair of lists
;(compile 'vectorized-and)

(defun non-nil (list) 
  (remove-if #'(lambda (x) (eq x nil)) list) )
;(compile 'non-nil)

(defun outer-list (x y) (outer-product x y #'combine))
; yields an outer-product array of lists, each containing a pair of
; elements of arrays (or vectors, but not lists) x and y
;(compile 'outer-list)

(defun ravel (array)
; unravels the elements of an array into a single, flat list
	(let* ((dim (array-dimensions array))
		   (length (prod dim))
		   (vector (make-array length :displaced-to array))
		  )
		  (coerce vector 'list)))
;(compile 'ravel)

#|(defun matrix-rank (ssp-matrix)
; returns the rank of a non-negative-def symmetric matrix
	(let* ((e-values (eigenvalues ssp-matrix))
	      )
	      (length (which (> e-values 0.0000001)))
	      ))
|#
	      
(defun matrix-rank (square-matrix)
; an alternative approach based on the sweep operator
; returns the rank of any square matrix
; by seeing how many rows get swept
	(length (select 
	  (sweep-operator square-matrix (iseq (first (array-dimensions square-matrix))))
	           1)))
	      
(defun moore-penrose (matrix) ; notation as in the Mathematica book, p665
; returns the Moore-Penrose g-inverse of any square matrix 
; -- doesn't work for some reason (a bug) on any matrix that has at least as many 
; rows as columns
	(let* ((svd (sv-decomp matrix))
	       (sv (select svd 1))
	       (sv-indices (which (> (^ sv 2) 0.0000001)))
	       (row-indices (iseq (select (array-dimensions matrix) 0)))
	       (mdinv (diagonal (/ 1 (select sv sv-indices))))
	       (u (transpose (select (select svd 0) row-indices sv-indices)))
	       (v (select (select svd 2) row-indices sv-indices))
	      )
	    (matmult v mdinv u)))
	    
(defun moore-penrose-ssp (x &optional w)
; (this function seems to have a bug!)
; returns the moore-penrose inverse of (x^T)diag(w)x
; where x is a general n by p matrix, possibly with
; rank less than p, and w is a list of weights.
; If w unspecified, it is taken as a list of ones.
; Uses methods outlined in Wetherill's "regression analysis".
	(let* ((n (array-dimension x 0))
		   (p (array-dimension x 1))
	       (w (if w (sqrt w) (repeat 1 n)))
	       (x (matmult (diagonal w) x))
	       (xt (transpose x))
	       (xxt (cross-product xt))
	       (xxt-inv (inverse xxt))
	      )
	      (matmult xt xxt-inv xxt-inv x)
	)) 
 
	    
(defun ginv (matrix) ; returns a g-inverse of a square matrix
                     ; by use of the sweep operator
    (let* ((p (select (array-dimensions matrix) 0))
          )
          (select (sweep-operator matrix (iseq p)) 0)))
	      
(defun conform-to-shape (nested-list target-list)
; for equal-length lists (combine nested-list) and (combine target-list)
; this function results in a nested list with the same "shape"
; as nested-list, but with elements taken from target-list
	(cond ((= (length (combine target-list)) (length (combine nested-list)))
	         (cond ((flat-list-p nested-list) target-list)
	               ((not (compound-data-p nested-list)) target-list)
	               (t (let ((result '())
	                     (target-list (combine target-list)))
	                  (dolist (term nested-list result)
	                    (setf data-for-term 
	                       (select target-list (iseq (length (combine term)))))
	                    (setf target-list 
	                       (select target-list 
	                         (which (> (iseq (length target-list)) 
	                                (- (length (combine term)) 1)))))
	                    (setf result  
	                      (append result 
	                        (list (conform-to-shape term data-for-term)))))))))
	      (t (error
	           "Lists not conformable -- wrong number of elements")))
)
	
(defun flat-list-p (list)
; does the list contain only simple items?
(cond ((listp list)
       (reduce #'both (mapcar #'not (mapcar #'compound-data-p list)))
      )
      (t nil)))
      
(defun log-ratio (x y)
	(cond ((and (> x 0) (> y 0)) (log (/ x y)))
	       (t -99)))
;(compile 'log-ratio)

(defun poisson-deviance (o e)
	(* 2 (- (inner-product o (map-elements #'log-ratio o e)) (sum (- o e)))))
;(compile 'poisson-deviance)

(defun print-without-parens (list)
; a string representation of the list without the enclosing parentheses
(strip-first-and-last (with-output-to-string (s) (prin1 list s))))

(defun strip-first-and-last (string)
; strips the first and last characters from a string
(concatenate 'string (mapcar #'(lambda (index) (char string index)) 
                                  (iseq 1 (- (length string) 2)))))
                                  
(defun reduce-by-plus (list) (reduce #'+ list))
;(compile 'reduce-by-plus)

(defun multiply-elementwise (list1 list2) (map-elements #'* list1 list2))
;(compile 'multiply-elementwise)

(defun table-print (x defining-factors i j &key (float-digits 3))
"Args: array
Sort of pretty-prints any array X. Makes slices
along dimensions I and J. If X is 10 x 3 x 7,
then (array-print x 0 1) prints 7 matrices of
dimensions 10 x 3, while (array-print x 2 0)
prints 3 matrices of dimensions 7 x 10."
(let* ((i (cond ((vectorp x) (+ 1 i)) (t i)))
       (j (cond ((vectorp x) (+ 1 j)) (t j)))
       (x (cond ((vectorp x) (make-array (list 1 (length x)) :initial-contents x)) 
         (t x)))
       (mx (array-dimensions x))
       (nx (array-total-size x))
       (i-name (cond ((< i 0) nil) (t (select defining-factors i))))
       (j-name (last defining-factors))
      ) 
(dotimes (k nx) 
(let* (
       (sk (array-subscript x k))
       (si (cond ((< i 0) 0) (t (elt sk i))))
       (sj (cond ((> j (- (length sk) 1)) (last sk))   (t (elt sk j)) ))
      )
(if (and (= 0 si) (= 0 sj))
(progn
 (cond ((< i 0) nil)
       (t (setf (elt sk i) i-name)))
 (cond ((> j (- (length sk) 1)) nil) (t (setf (elt sk j) j-name)))
 (cond ((= (first mx) 1) (list (put-together (second sk) (first defining-factors))))
       (t (map-elements #'put-together sk defining-factors)))
 ;(print-out sk defining-factors)
 ;(print (coerce sk 'vector))
 (terpri)
 (setf (elt sk i) nil)
 (cond ((> j (- (length sk) 1)) nil) (t (setf (elt sk j) nil)))
 (print-matrix 
    (cond ((= (first mx) 1) (make-array (list 1 (second mx)) 
                              :initial-contents (array-slice x sk)))
          (t (array-slice x sk)))
   t :float-digits float-digits)
 (format nil "")
))))))

(defun put-together (sk factor-name)
	(cond ((numberp sk) (format t "[~a~a~d~a" (symbol-name factor-name) " = " 
	                        (+ 1 sk) "]  "))
	      (t            (format t "[~a] " (symbol-name factor-name)))))
	      
(defun submatrix (matrix ilist jlist) ; extracts a submatrix of a matrix
	(setf result nil)
	(dolist (i ilist)
	        (dolist (j jlist)
	                (setf result (append result (list (select matrix i j))))))
	(make-array (list (length ilist) (length jlist)) :initial-contents result))

(defun index-betas (betas) ; maps columns of the design matrix to betas
	   (do ((i 0 (+ i 1))
	        (result nil (append result
	           (cond ((compound-data-p (first (select betas i)))
	                 (do ((j 0 (+ j 1))
	                      (subresult nil (append subresult
	                       (repeat (+ (* 10 i) j) 
	                        (length (select (select betas i) j))))))
	                    ((> j (- (length (select betas i)) 1)) subresult)))
	              (t (repeat (* 10 i) (length (select betas i))))))))
	     ((> i (- (length betas) 1)) result)))
	     
	    
	        
