
(terpri)
(format t "LLAMA version 0.15, 21 August 98, copyright (c) David Firth")
(terpri)

(require "array")
(require "ll-macs")
(require "ll-util")
(require "ll-modl")
(require "ll-tabl")
(require "ll-menu")
(require "ll-prin")
(terpri)
     

(defun deviance (o e) (poisson-deviance o e))

(defun modelp (object-name)
; determines whether a symbol is the name of a model
	(cond ((boundp object-name) 
	          (cond ((objectp (eval object-name)) 
	                   (cond ((eq (send (eval object-name) 
	                       :slot-value 'proto-name) 'model-proto) t)
	                         (t nil)))
	                (t nil)))
	      (t nil)))

(defun tablep (object-name)
; determines whether a symbol is the name of a table object
	(cond ((boundp object-name) 
	          (cond ((objectp (eval object-name)) 
	                   (cond ((eq (send (eval object-name) 
	                       :slot-value 'proto-name) 'table-proto) t)
	                         (t nil)))
	                (t nil)))
	      (t nil)))
	      
(defun design-element (term-subsets term-betas)
; returns the implicit design-matrix columns for a single model term
	(cond ((compound-data-p (first term-betas)); is the term multiplicative?
			(reduce #'bind-columns
			(mapcar #'(lambda (j) (item-design-element j term-subsets term-betas))
			   (iseq 1 (length term-betas)))))
		  (t 
		  		(reduce #'bind-columns term-subsets)
		  )
	)
)

(defun item-design-element (j term-subsets term-betas)
; returns the design columns for the jth item of a multiplicative term
; (typically out of two items? or more generally?)
	(let* ((length (length term-betas))
		   (subterm-betas (select term-betas (- j 1))) ; a simple list
		   (subterm-subsets (select term-subsets (- j 1)))
		   	       (other-betas (select term-betas (remove (- j 1) (iseq length))))
	       (other-betas (select term-betas (remove (- j 1) (iseq length))))
	       (other-subsets (select term-subsets (remove (- j 1) (iseq length))))
	       (derivative-factors 
	          (mapcar #'reduce-by-plus
	              (map-elements #'(lambda (beta-list subset-list)
	                (map-elements #'(lambda (beta subset)
	                            (* beta subset)) beta-list subset-list))
	                           other-betas other-subsets)))
	       (derivative 
	          (cond ((compound-data-p derivative-factors)
	                  (reduce #'* derivative-factors))
	                 (t
	                  derivative-factors)))
	       (columns 
	         (mapcar #'(lambda (subset) (* subset derivative)) subterm-subsets))
	      )
	      (reduce #'bind-columns columns)
	      ))
	         
(defun expand-factors (factor-list)   ; no longer needed
; calculates values for each factor in an ordered list
; and puts them in the appropriate "slot" in the factor list
	(let* ((ways (length factor-list))
	       (dimensions (mapcar #'get-levels factor-list))
		   (table-size (prod dimensions))
		   (names (mapcar #'get-name factor-list))
		   (cumulative-products (combine 1 (accumulate #'* dimensions)))
		   (temp '())
		   (values 
		     (dotimes (i ways temp)
		          (setf temp (append temp (list
		             (let* ((dimension (select dimensions i))
		               (cum-prod (select cumulative-products (+ i 1)))
		               (cum-prod-lagged (select cumulative-products i))
		              )
		              (repeat 
		                (repeat (iseq 1 dimension) 
		                  (repeat (/ table-size cum-prod) dimension))
		                cum-prod-lagged))
		            )))))
		  )
		  (map 'list #'list names dimensions values)))
		  
(defun dimensions (model-or-table 
  &optional (simple-term (send model-or-table :factors) set))
; returns a list containing the dimensions of a simple linear or interaction term
; in a model or a table
	(if set (mapcar #'(lambda (factor-name) 
	    (get-levels-by-name model-or-table factor-name))
	         simple-term)
	   (mapcar #'factor-levels simple-term))
)
;(compile 'dimensions)

(defun get-levels-by-name (model-or-table factor-name)
; returns the number of levels for a factor with a given name
	(let* ((factors (send model-or-table :factors))
	       (names (mapcar #'factor-name factors))
	       (dimensions (mapcar #'factor-levels factors))
	      )
	      (first (select dimensions 
	         (which (map-elements #'eq names factor-name))))))

(defun fit-function (model-object-name &key (iterations 10))
	(setf *current-model* model-object-name)
	(let* ((model-object (eval model-object-name)))
	(cond ((not (send model-object :terms)) 
	        (format t "No can do: The model has no terms specified") nil)
	      ((not (send model-object :counts)) 
	        (format t "No can do: The table has no counts specified") nil)
	      ((not (terms-are-all-factors-p model-object))
	        (format t "No can do: Model terms involve undefined factors") nil)
	      (t (fit-routine model-object-name iterations))))
)

(defun fit-routine (model-object-name iterations)
	(terpri)
	(format t "Fitting the model ~a ... please be patient" 
	            (symbol-name model-object-name) nil)
	(let* ((model-object (eval model-object-name))
	       (counts (send model-object :counts))
		   (factors (send model-object :factors))
		   (terms (send model-object :terms))
		   (count-subsets (mapcar #'(lambda (term) 
		        (make-count-subsets term factors)) terms))
		   (count-margins (mapcar #'(lambda (term-subsets)
		        (make-count-margins term-subsets counts)) count-subsets))
		   (betas ; perhaps tidy this up to avoid repetition of code
		          ; by using a lambda-function?
		      (flet ((term-to-betas (term)
		        (cond ((compound-data-p (first term)) ;is the term multiplicative?
		      		     (flet ((mult-to-betas (subterm)
		      		     	(let* 
		                   ((dimensions 
		                      (mapcar #'(lambda (factor-name)
		                 (get-levels-by-name model-object factor-name))
		                         subterm))
		                   )
		                   (+ 1 (* (normal-rand (prod dimensions)) 0.05)))); 
		                       )
		                       (mapcar #'mult-to-betas term)) 
		      		  )
		               (t                            ;or is it just a linear term?
		                 (let* 
		                   ((dimensions 
		                      (mapcar #'(lambda (factor-name)
		                 (get-levels-by-name model-object factor-name))
		                         term))
		                   )
		                   (+ 1 (* (normal-rand (prod dimensions)) 0.05)))))); 
		             )
		             (mapcar #'term-to-betas terms)))
		   (effects (map-elements #'effect count-subsets betas))
		  )
		  (send model-object :betas betas)
		  (send model-object :subsets count-subsets)
		  ;(send model-object :fit-values (repeat 1 (length counts)))
		  (send model-object :effects effects)
		  (send model-object :calculate-fit-values)
		  (send model-object :count-margins count-margins)
		  (send model-object :needs-computing nil)
		  (ipf model-object iterations)
		  (send (eval (send model-object :data-table))
		      :add-model model-object-name)
		  (terpri)
		  (format t "...for a further n iterations, use (RECYCLE n)..." 
		     nil)
		  (values)
	)
)
;(compile 'fit-routine)

(defun recycle (&optional (iterations 1) 
                          (model-object-name *current-model*))
	(terpri)
	(format t "Iterating again on model ~a ... please be patient" 
	            (symbol-name model-object-name) nil)
	(ipf (eval model-object-name) iterations)
	(values)
)

(defun ipf (model-object iterations) ; returns betas
	(do* ((printout nil (terpri))
	     (printout nil (princ "iteration "))
	     (iteration 0 (princ (+ 1 iteration)))
		 (betas nil
		     (update model-object))
		 (printout nil (princ ", deviance "))
		 (deviance nil 
		           (princ 
		             (deviance (send model-object :counts)
		                       (send model-object :fit-values))))
		)
		((or (= iteration iterations) (converged-p)) betas)
	  )
	
)

(defun converged-p nil)

(defun update (model-object) ; returns new betas
  ;(print "update")
  (let* ((count-subsets (send model-object :subsets))
         (count-margins (send model-object :count-margins))
         (betas (send model-object :betas))
         (exposures (send model-object :exposures))
         (counts (send model-object :counts))
         (fit-values (send model-object :fit-values))
         (factors (send model-object :factors))
         (terms (send model-object :terms))
        )
  (dolist (i (iseq 1 (length terms)) betas) ; do for each term
     (let* ((term-subsets (select count-subsets (- i 1)))
            (term-margins (select count-margins (- i 1)))
            (term-old-betas (select betas (- i 1)))
            (fit-values (send model-object :fit-values))
            (effects (send model-object :effects))
            (term-old-effect (select effects (- i 1)))
            (term (select terms (- i 1)))
            (term-new-betas 
	              (cond ((compound-data-p (first term)) ;is it a multiplicative term?
	                     (setf updated-term '())
	                     (dolist (j (iseq 1 (length term)) updated-term) ; for each subterm
	                      (setf updated-term (append updated-term
	                        (list (update-subterm 
	                           j term-old-betas term-subsets counts fit-values))))))
	                    (t 
	                      (update-term term-old-betas term-subsets term-margins 
	                          counts fit-values))
	               )
	      	)
            (term-new-effect (effect term-subsets term-new-betas))
            (fit-values (* fit-values (/ term-new-effect term-old-effect)))
           )
        ;(break)
        (setf (select (send model-object :effects) (- i 1)) term-new-effect)
        (send model-object :fit-values fit-values)
        (setf (select betas (- i 1)) term-new-betas) 
             ))))

(defun iwls-update (model-object) 
; updates the complete set of betas
  (cond ((loglinear-model-p model-object); is the model loglinear?
	(let* ((fitted (send model-object :calculate-fit-values))
	       (eta (log fitted))
	       (counts (send model-object :counts))
	       (betas (send model-object :betas))
	       (beta-vector (coerce (combine betas) 'vector))
	       (w (diagonal fitted))
	       (design-matrix (send model-object :local-design-matrix))
	       (lp (matmult design-matrix beta-vector))
	       (wu (coerce (/ (- counts fitted) fitted) 'vector))
	       (z (+ lp wu))
	       (p (select (array-dimensions design-matrix) 1))
	       #|(g-inverse 
	         (select 
	           (sweep-operator 
	           (matmult (transpose design-matrix) w design-matrix) (iseq p))
	         0))|#
	       (g-inverse (moore-penrose 
	          (matmult (transpose design-matrix) w design-matrix)))
	       (new-betas (matmult g-inverse (transpose design-matrix) w z))
	       ;(d-beta (matmult g-inverse (transpose design-matrix) (- counts fitted)))
	      )
	      (send model-object :betas (conform-to-shape betas 
	               new-betas))
	      (send model-object :calculate-deviance)
	   ))
	   (t   ; if the model is log-multiplicative
	      (twostep-iwls-update model-object)
	   )
	 )
)
	 
#|(defun iwls-update (model-object) 
; updates the complete set of betas
	(let* ((fitted (send model-object :calculate-fit-values))
	       (eta (log fitted))
	       (counts (send model-object :counts))
	       (exposures (send model-object :exposures))
	       (betas (send model-object :betas))
	       (beta-vector (coerce (combine betas) 'vector))
	       (w (diagonal fitted))
	       (x (send model-object :local-design-matrix))
	       (x-transpose (transpose x))
	       (count-subsets (send model-object :subsets))
	       (terms (send model-object :terms))
	       (factors (send model-object :factors))
	       ;(lp (matmult x beta-vector))
	       ;(wu (coerce (/ (- counts fitted) fitted) 'vector))
	       (u (coerce (- counts fitted) 'vector))
	       ;(z (+ lp wu))
	       (p (select (array-dimensions x) 1))
	       (g-inverse 
	         (select 
	           (sweep-operator 
	           (matmult x-transpose w x) (iseq p))
	         0))
	       #|(g-inverse (moore-penrose 
	          (matmult x-transpose w x)))|#
	       (new-eta (+ eta (matmult
	          x g-inverse x-transpose u)))
	       (new-fit (exp new-eta)) ; no longer in the model
	       (new-betas (update betas new-fit exposures 
	             factors terms count-subsets))
	      )
	      (send model-object :betas new-betas)
	      (print (send model-object :calculate-deviance))
	      ))|#
	      
(defun twostep-iwls-update (model-object)
	(let* ((first-model (first-model model-object))
	       (first-model-betas (iwls-update first-model))
	       (second-model (second-model model-object first-model-betas))
	       (second-model-betas (iwls-update second-model))
	       (model-betas 
	          (combine-betas model-object first-model-betas second-model-betas))
	      )
	      (send model-object :betas model-betas)))
	      
(defun first-model (model-object)
	(let* ((terms (send model-object :terms))
	       (terms (mapcar #'convert-first terms))
	      )
	      nil))
	      
(defun loglinear-model-p (model)
; is the model loglinear?
; (alternative, a log-multiplicative model, returns nil)
	(reduce #'both (mapcar #'flat-list-p (send model :terms))))
	      
(defun update-subterm (j term-betas term-subsets counts fit-values)
	;(print "update-subterm")
	(let* ((length (length term-betas))
	       (subterm-betas (select term-betas (- j 1))) ; a simple list
	       (subterm-subsets (select term-subsets (- j 1)))
	       (other-betas (select term-betas (remove (- j 1) (iseq length))))
	       (other-subsets (select term-subsets (remove (- j 1) (iseq length))))
	       (derivative-factors 
	          (mapcar #'reduce-by-plus
	              (map-elements #'(lambda (beta-list subset-list)
	                (map-elements #'(lambda (beta subset)
	                            (* beta subset)) beta-list subset-list))
	                           other-betas other-subsets)))
	       (derivative 
	          (cond ((compound-data-p derivative-factors)
	                  (reduce #'* derivative-factors))
	                 (t
	                  derivative-factors)))
	      )
	      (map-elements #'(lambda (beta subset) 
	        (let* ((subset-by-deriv (* subset derivative))
	               (count-margin (inner-product counts subset-by-deriv))
	               (fit-by-subset-by-deriv (* fit-values subset-by-deriv))
	               (fitted-margin (sum fit-by-subset-by-deriv))
	               (denominator (inner-product fit-by-subset-by-deriv derivative))
	              )
	        (+ beta (/ (- count-margin fitted-margin) denominator))))
	         subterm-betas subterm-subsets)))
	         
	         
(defun update-subterm (j term-betas term-subsets counts fit-values)
	(let* ((length (length term-betas))
	       (subterm-betas (select term-betas (- j 1))) ; a simple list
	       (subterm-subsets (select term-subsets (- j 1)))
	       (other-betas (select term-betas (remove (- j 1) (iseq length))))
	       (other-subsets (select term-subsets (remove (- j 1) (iseq length))))
	       (derivative-factors 
	          (mapcar #'reduce-by-plus
	              (map-elements #'(lambda (beta-list subset-list)
	                (map-elements #'(lambda (beta subset)
	                            (* beta subset)) beta-list subset-list))
	                           other-betas other-subsets)))
	       (derivative 
	          (cond ((compound-data-p derivative-factors)
	                  (reduce #'* derivative-factors))
	                 (t
	                  derivative-factors)))
	      )
	      (map-elements #'(lambda (beta subset) 
	        (let* ((subset-by-deriv (* subset derivative))
	               (count-margin (inner-product counts subset-by-deriv))
	               (fit-by-subset-by-deriv (* fit-values subset-by-deriv))
	               (fitted-margin (sum fit-by-subset-by-deriv))
	               (denominator (inner-product fit-by-subset-by-deriv derivative))
	              )
	        (+ beta (log (+ 1 (/ (- count-margin fitted-margin) denominator))))))
	         subterm-betas subterm-subsets)))
	         
(defun update-term (term-betas term-subsets term-margins counts fit-values)
     ;(print "update-term")
     (let* ((fitted-margins 
               (mapcar #'(lambda (subset) (inner-product fit-values subset)) term-subsets))
           )
     (+ term-betas (map-elements #'log-ratio term-margins fitted-margins)))
)

(defun effect (term-subsets term-betas) ;returns multiplicative effect of a term
	;(print "effect")
	(cond ((compound-data-p (first term-betas)) ;is it a multiplicative term?
	          (exp (reduce #'*
	                 (mapcar #'reduce-by-plus
	                     (map-elements #'multiply-elementwise 
	                          term-subsets term-betas))
	               )
	          )
	      )
	      (t 
	         (exp (reduce #'+ (map-elements #'* term-subsets term-betas)))
	               )))
	               
(defun make-count-margins (term-subsets counts)
; calculates all the count margins for a single term
; -- called in fit-routine
	(cond ((compound-data-p (first (first term-subsets))); multiplicative term?
	       (mapcar #'(lambda (subsets) (make-count-margins subsets counts))
	                term-subsets))
	      (t ; if a "flat" term
	        (mapcar #'(lambda (subset) (inner-product subset counts)) term-subsets))))

(defun make-count-subsets (term factors)
	; returns a list of indicator vectors corresponding to 
	; level combinations in term
	;(print "make-count-subsets")
	(cond ((compound-data-p (first term)) ;is it a multiplicative term?
	       (mapcar #'(lambda (sub-term)
	         (mapcar #'as-binary
	           (let* ((factors-in-subterm (flet ((pull-factor (fac-name)
		                                (extract-factor fac-name factors)))
		                             (mapcar #'pull-factor sub-term)))
		            (subsets '())
		           )
		        (dolist (i 
            (ravel (reduce #'outer-list 
                    (mapcar #'get-levels-as-vector factors-in-subterm)))
            subsets)
	        (def subsets 
	          (append subsets (list
	            ;select all counts with combination i of factors
	            (reduce #'vectorized-and 
	                  (map-elements #'= (mapcar #'factor-values 
	                          factors-in-subterm) i)))))))
	          )
	                )
	         term)
	         
	         
	        )
	(t
	  (mapcar #'as-binary 
	    (let* ((factors-in-term (flet ((pull-factor (fac-name)
		                                (extract-factor fac-name factors)))
		                             (mapcar #'pull-factor term)))
		   (subsets '())
		  )
		(dolist (i 
            (ravel (reduce #'outer-list 
                    (mapcar #'get-levels-as-vector factors-in-term)))
            subsets)
	        (def subsets 
	          (append subsets (list
	            ;select all counts with combination i of factors
	            (reduce #'vectorized-and 
	                  (map-elements #'= (mapcar #'factor-values factors-in-term) i)))
	          ))))))))


    




    
